; 
; File:         ConverterDF.inc
; Written by:   RuiLoureiro         @masmforum
;--------------------------------------------------------------------------
;  To get information about these procedures, read the files 
;
;           Converter4.txt  or   CnvRealDD.txt
;
; Use:
;       invoke  ConvertString4D, addr String, addr DwordVar
;
; -----
;       invoke  ConvertFloat4DF, addr DwordVar, addr String
;
;              where String must be defined with a DWORD behind
;
;                              dd 0           <<<<<<- for length
;                    String    db x dup(?) 
;  OR
;
;       invoke  ConvertReal4DF, addr DwordVar, addr String 
;
;              where 
;                    String    db x dup(?) 
; 
ConvertString4D     proto   :DWORD,:DWORD
ConvertFloat4DF     proto   :DWORD,:DWORD
ConvertReal4DF      proto   :DWORD,:DWORD

$DECIMALPLACES_REAL4  equ 6     ; for  ConvertFloat4DF


$INTEGERDIGITS_REAL4  equ 8     ; no more than 8 integer digits
$DECIMALDIGITS_REAL4  equ 7     ; and 7 decimals
$EXPONENTDIGITS_REAL4 equ 2     ; and 2 for exponent

;----------------------------------
;        characteres
;----------------------------------
$PLUS$          equ 2Bh
$MINUS$         equ 2Dh
$ZERO$          equ 30h
$SPACE$         equ 20h
;----------------------
$SPACE_ZERO$    equ -16         ; $SPACE$ - $ZERO$
$PLUS_ZERO$     equ -5
$MINUS_ZERO$    equ -3
$NULL_ZERO$     equ -48
$POINT_ZERO$    equ -2

$DIGIT_ZERO$    equ 9           ; limit is 9: from 0 to 9
$E_ZERO$        equ 21
$e_ZERO$        equ 53
; -----------------------------------
$AND_ROUND$    equ 0f3ffh     ; code it for rounding
$OR_TRUNCATE$  equ 0C00h      ; code it for truncating
$TRUNCATE$     equ 027fh
; 
;                                   macros
; 
MULBY10ADDEAX       MACRO   iReg
                    lea     iReg, [iReg+iReg*4]
                    lea     iReg, [eax+iReg*2]
ENDM
;-----------------------------------------------------------------
ADDDECIMALPOWERDD   MACRO   iregS, iregE
                    LOCAL   getout
                    
                    fld     qword ptr [_PowerTblDQM38+iregS*8]
                    fmul
                    fadd
                    
                    or      iregE, iregE
                    jz      short getout
                    js      short @F

                    fld     qword ptr [_PowerTblDQP38+iregE*8]
                    fmul
                    jmp     getout
                    
            @@:     neg     iregE
                    fld     qword ptr [_PowerTblDQM38+iregE*8]
                    fmul
        getout:
ENDM
; 
EXAMINE2REAL4   MACRO

                mov     eax, dword ptr [eax]
                or      eax, eax
                jz      _finish
                                                
                mov     edx, eax
                shr     edx, 23         ; exponent and sign in DX; DH=sign DL=exponent
                and     eax, 07FFFFFH   ; fraction in EAX
                ;
                cmp     dl, 0FFh        ; FFh => NAN or INFINITY
                jne     short _finish
                        
                or      eax, eax
                jz      _erro8          ; overflow / underflow
                ; --------------------
                ; Indefinite or NAN
                ; --------------------
                jmp     _erro9
ENDM
; 
EXAMINE3REAL4       MACRO
                    LABEL   notNAN, finish0
                    
                    mov     edx, dword ptr [eax]
                                                
                    mov     eax, edx
                    shr     edx, 23             ; exponent and sign in DX: DH=sign (bit 0), DL=exponent

                    and     eax, 07FFFFFH       ; fraction in EAX  <<<<<<<<<<<<<<<<<<< NEW HERE
                    
                    cmp     dl, 0FFh            ; FFh => NAN or INFINITY
                    jne     short notNAN        ; not NAN or infinity                    

                    ; -----------------------
                    ;  It is NAN or INFINITY
                    ; -----------------------
                    or      eax, eax
                    jz      _erro8              ; it is +INFINITY or -INFINITY
                    ;jz      short finish0      ; it is +INFINITY or -INFINITY
                    jnz     _erro9 

                    ;------------------------------------------------
                    ;           It is not NAN or INFINITY
                    ; DL  = exponent
                    ; DH  = sign ( bit 0 )
                    ; EAX = fraction bits
                    ;------------------------------------------------
        notNAN:     test    dl, dl
                    jnz     short finish0        ; exponent not FFh and not 0

                    ;---------------------------------------------
                    ; exponent=0, if fraction=0 => it is +0 or -0
                    ; if frction not 0 => denormalized
                    ;---------------------------------------------
                    or      eax, eax
                    jnz     _erro9              ; denormalized

                    ;------------------------------------------------
                    ;                 EXIT HERE
                    ;------------------------------------------------
       finish0:              
ENDM
; ----------------------------------------------------------------------------------------------
ERRORS3TO9      MACRO

                ; ---------------------------
                ; more than 8 integer digits
                ; ---------------------------
    _erro3:     mov     eax, 3
                mov     edx, offset _StrError3
                jmp     _error

                ; ---------------------------
                ; more than 7 decimal digits
                ; ---------------------------
    _erro4:     mov     eax, 4
                mov     edx, offset _StrError4
                jmp     _error

                ; --------------------------
                ; more than 2 exponent digits
                ; --------------------------
    _erro5:     mov     eax, 5
                mov     edx, offset _StrError5
                jmp     _error

                ; ---------------------------
                ;    undefined pointer
                ; ---------------------------
    _erro6:     mov     eax, 6
                mov     edx, offset _StrError6
                jmp     _error

                ; ---------------------------
                ;   FPU invalid operation
                ; ---------------------------
    _erro7:     fstp    st(0)                   ; remove st(0)
                fclex
                mov     eax, 7
                mov     edx, offset _StrError7
                jmp     _error

                ; ----------------------
                ;  overflow / underflow
                ; ----------------------
    _erro8:     mov     eax, 8
                mov     edx, offset _StrError8
                jmp     _error

                ; --------------------------
                ;    Indefinite or NAN
                ; --------------------------
    _erro9:     mov     eax, 9
                mov     edx, offset _StrError9
                jmp     _error
ENDM

; 
EXPONENTREAL_4W MACRO
                
                mov     cl, 10
                
                mov     bx, '+E'
                mov     eax, exponent

                or      eax, eax
                jns     short @F
                
                mov     bx, '-E'
                neg     eax                     ; make number positive
        @@:      
                div     cl              
                or      ax, 3030h
                shl     eax, 16
                or      ax, bx
                mov     [edi], eax
                add     edi, 4
ENDM
; 
EXPONENTREAL_4R MACRO
                
                mov     cl, 10
                
                mov     bx, '+E'
                
                or      eax, eax
                jns     short @F
                
                mov     bx, '-E'
                neg     eax                     ; make number positive
                
        @@:     div     cl              
                or      ax, 3030h
                shl     eax, 16
                or      ax, bx
                mov     [edi], eax
                add     edi, 4
ENDM
; 
DIVISION2BY10   MACRO
                mov     edx, 3435973837                
                mov     ecx, eax
                mul     edx                 ; gives EDX:EAX
                shr     edx, 3

                mov     eax, edx
                lea     edx, [edx+edx*4]
                add     edx, edx

                sub     ecx, edx
                or      cl, 30h     ;'0'
ENDM
; 
SETTRUNCATE         MACRO    oldcw, truncw
                    fstcw    oldcw
                    fwait                     
                    mov      word ptr truncw, $TRUNCATE$
                    fldcw    truncw
ENDM
; 
.data
                dd 16
_StrError1      db "expression error"
                dd 22
_StrError2      db "unacceptable character"

                dd 26
_StrError3      db "more than 8 integer digits"
                dd 26
_StrError4      db "more than 7 decimal digits",0
                dd 27
_StrError5      db "more than 2 exponent digits",0

                dd 17
_StrError6      db "undefined pointer",0

                dd 21
_StrError7      db "FPU invalid operation",0
                dd 21
_StrError8      db "overflow or underflow",0
                dd 24
_StrError9      db "indefinite or NAN number",0

;-------------------------------------------------------------------------
_PowerTblDQP38  dq 1.0                              ; 0
                dq 10.0                             ; 1
                dq 1.0e2                            ; 2
                dq 1.0e3                            ; 3
                dq 1.0e4                            ; 4
                dq 1.0e5                            ; 5
                dq 1.0e6                            ; 6
                dq 1.0e7                            ; 7
                dq 1.0e8                            ; 8
                dq 1.0e9                            ; 9
                dq 1.0e10                           ; 10
                ;
                dq 1.0e11                           ; 11
                dq 1.0e12                           ; 12
                dq 1.0e13                           ; 13
                dq 1.0e14                           ; 14
                dq 1.0e15                           ; 15
                dq 1.0e16                           ; 16
                dq 1.0e17                           ; 17
                dq 1.0e18                           ; 18
                dq 1.0e19                           ; 19
                dq 1.0e20                           ; 20
                ;
                dq 1.0e21                           ; 21
                dq 1.0e22                           ; 22
                dq 1.0e23                           ; 23
                dq 1.0e24                           ; 24
                dq 1.0e25                           ; 25
                dq 1.0e26                           ; 26
                dq 1.0e27                           ; 27
                dq 1.0e28                           ; 28
                dq 1.0e29                           ; 29
                dq 1.0e30                           ; 30
                ;
                dq 1.0e31                           ; 31
                dq 1.0e32                           ; 32
                dq 1.0e33                           ; 33
                dq 1.0e34                           ; 34
                dq 1.0e35                           ; 35
                dq 1.0e36                           ; 36
                dq 1.0e37                           ; 37
                dq 1.0e38                           ; 38
                dq 1.0e39                           ; 39
                dq 1.0e40                           ; 40
                ;
                dq 1.0e41                           ; 41
                dq 1.0e42                           ; 42
                dq 1.0e43                           ; 43
                dq 1.0e44                           ; 44
                dq 1.0e45                           ; 45
                dq 1.0e46                           ; 46
                

_PowerTblDQM38  dq 1.0                              ; 0
                dq 0.1                              ; 1
                dq 1.0e-2                           ; 2
                dq 1.0e-3                           ; 3
                dq 1.0e-4                           ; 4
                dq 1.0e-5                           ; 5
                dq 1.0e-6                           ; 6
                dq 1.0e-7                           ; 7
                dq 1.0e-8                           ; 8
                dq 1.0e-9                           ; 9
                dq 1.0e-10                          ; 10
                ;
                dq 1.0e-11                          ; 11
                dq 1.0e-12                          ; 12
                dq 1.0e-13                          ; 13
                dq 1.0e-14                          ; 14
                dq 1.0e-15                          ; 15
                dq 1.0e-16                          ; 16
                dq 1.0e-17                          ; 17
                dq 1.0e-18                          ; 18
                dq 1.0e-19                          ; 19
                dq 1.0e-20                          ; 20
                ;
                dq 1.0e-21                          ; 21
                dq 1.0e-22                          ; 22
                dq 1.0e-23                          ; 23
                dq 1.0e-24                          ; 24
                dq 1.0e-25                          ; 25
                dq 1.0e-26                          ; 26
                dq 1.0e-27                          ; 27
                dq 1.0e-28                          ; 28
                dq 1.0e-29                          ; 29
                dq 1.0e-30                          ; 30
                ;
                dq 1.0e-31                          ; 31
                dq 1.0e-32                          ; 32
                dq 1.0e-33                          ; 33
                dq 1.0e-34                          ; 34
                dq 1.0e-35                          ; 35
                dq 1.0e-36                          ; 36
                dq 1.0e-37                          ; 37
                dq 1.0e-38                          ; 38
                dq 1.0e-39                          ; 39
                dq 1.0e-40                          ; 40
                ;
                dq 1.0e-41                          ; 41
                dq 1.0e-42                          ; 42
                dq 1.0e-43                          ; 43
                dq 1.0e-44                          ; 44
                dq 1.0e-45                          ; 45
                dq 1.0e-46                          ; 46                
; ----------------------------------------------------------------------------------------------
.data?
_DWinteger          dd ?

.code
; 
ConvertString4D     proc    pStr:DWORD, pOut4:DWORD
                    LOCAL   flagZI      :DWORD
                    LOCAL   flagZE      :DWORD                    
                    LOCAL   IntegerSign :DWORD
                    LOCAL   ExponentSign:DWORD
                    LOCAL   integerDD   :DWORD
                    LOCAL   decimalDD   :DWORD
                                        
                    push    ebx
                    push    esi
                    push    edi
                    
                    mov     esi, pStr

                    mov     IntegerSign, 1              ; sign is minus
                    mov     ExponentSign, 1             ; sign is minus
                    mov     flagZI, 0                   ; =1 if we remeved 0
                    mov     flagZE, 0                   ; =1 if we removed 0
                    
                    mov     ecx, $INTEGERDIGITS_REAL4   ; number of integer digits

                    mov     integerDD, 0                ; integer part
                    xor     ebx, ebx                    ; number of decimal digits
                    xor     edi, edi                    ; decimal part
                    xor     edx, edx                    ; exponent
                        
                    ; --------------------------
                    ;     remove spaces
                    ; --------------------------
            @@:     movzx   eax, byte ptr [esi]
                    add     esi, 1
                    cmp     al, $SPACE$
                    je      short @B
                         
                    ; --------------------------
                    ;     Is it a sign ?
                    ; --------------------------
                    cmp     al, $MINUS$
                    je      short @F                    ; next may be spaces/zeroes

                    mov     IntegerSign, 0              ; is plus 
                                       
                    cmp     al, $PLUS$
                    je      short @F                    ; next may be spaces/zeroes
                    jne     short _testzeroI                    
                    
                    ; -------------------------
                    ;   it cannot have sign
                    ;  remove zeroes or spaces
                    ; -------------------------
    _removeZI:      mov     flagZI, 1
                    
            @@:     movzx   eax, byte ptr [esi]
                    add     esi, 1

    _testzeroI:     cmp     al, $ZERO$
                    je      short _removeZI
                    
                    cmp     al, $SPACE$
                    je      short @B                    
                    jne     short _isinteger 

            ;------------------------------------
            ;  It may be an integer or decimal
            ;------------------------------------
            
    _calcinteger:   sub     ecx, 1
                    js      _erro3
   
                    MULBY10ADDEAX ebx
      
    _nextinteger:   movzx   eax, byte ptr [esi]
                    add     esi, 1
                    
    _isinteger:     sub     al, $ZERO$ 
                    jc      short _isspaceI
                    
                    ; ----------------------------------
                    ; It must be digit, E, null or error2
                    ; ----------------------------------
                    cmp     al, $DIGIT_ZERO$
                    jbe     short _calcinteger
                    ;
                    mov     integerDD, ebx
                    xor     ebx, ebx                ; number of decimal digits                    
                    jz      _isexponentI            ; must be E, null or error2 

                    ; -------------------------------
                    ; It must be point, space or null
                    ; -------------------------------
    _isspaceI:      cmp     al, $SPACE_ZERO$        
                    je      short _nextinteger

                    ; --------------------------------
                    ; It must be null, point or error2
                    ; --------------------------------
                    mov     integerDD, ebx
                    xor     ebx, ebx                    ; number of decimal digits
                                        
                    cmp     al, $NULL_ZERO$        
                    je      _stopinteger

                    ; ----------------------------------
                    ;    It must be point or error2
                    ; ----------------------------------
                    cmp     al, $POINT_ZERO$
                    jne     _erro2                      ; unacceptable character

                    ; ------------------------
                    ; It may not have integers
                    ; ------------------------
                    mov     ecx, $DECIMALDIGITS_REAL4   ; number of decimal digits
                    je      short _nextdecimal

    _calcdecimal:   sub     ecx, 1
                    js      _erro4

                    add     ebx, 1
                    MULBY10ADDEAX edi

    _nextdecimal:   movzx   eax, byte ptr [esi]
                    add     esi, 1
                    
                    sub     al, $ZERO$ 
                    jc      short _isspaceD

                    ; ----------------------------------
                    ; It must be digit, E, null or error2
                    ; ----------------------------------
                    cmp     al, $DIGIT_ZERO$
                    jbe     short _calcdecimal
                    ja      _isexponentD

                    ; -------------------------------
                    ;   It must be space or null
                    ; -------------------------------
    _isspaceD:      cmp     al, $SPACE_ZERO$        
                    je      short _nextdecimal

                    cmp     al, $NULL_ZERO$        
                    je      _stopdecimal
                    jne     _erro2                      ; unacceptable character                    

                    ; -------------------------------
                    ;  It comes from integer part
                    ; Cannot start with E or +E or -E
                    ;    unless we removed zero
                    ; -------------------------------
    _isexponentI:   cmp     flagZI, 1
                    je      short @F
                    
                    cmp     ecx, $INTEGERDIGITS_REAL4
                    je      _erro1                      ; expression error

            @@:     mov     ecx, $EXPONENTDIGITS_REAL4   ; number of exponent digits

                    cmp     al, $E_ZERO$
                    je      short _startexponent
                    cmp     al, $e_ZERO$
                    je      short _startexponent
                    
                    ; ---------------------
                    ;   it may be null
                    ; ---------------------
                    cmp     al, $NULL_ZERO$        
                    je      _stopinteger
                    jne     _erro2                      ; unacceptable character                    

                    ; ----------------------------------
                    ;    It comes from decimal part
                    ;         cannot be .E 
                    ; ----------------------------------                    
    _isexponentD:   cmp     ecx, $DECIMALDIGITS_REAL4
                    je      _erro1
                    
                    mov     ecx, $EXPONENTDIGITS_REAL4   ; number of exponent digits                        
                    jne     short _startexponent

    _calcexponent:  sub     ecx, 1
                    js      _erro5
   
                    MULBY10ADDEAX edx

    _nextexponent:  movzx   eax, byte ptr [esi]
                    add     esi, 1
                    
    _isexponent:    sub     al, $ZERO$ 
                    jc      short _isspaceE

                    ; -------------------------
                    ; It must be digit or error
                    ; -------------------------
                    cmp     al, $DIGIT_ZERO$
                    jbe     short _calcexponent
                    ja      _erro2                      ; unacceptable character

                    ; ---------------------------------
                    ; It must be space or null or error2
                    ; ---------------------------------
    _isspaceE:      cmp     al, $SPACE_ZERO$        
                    je      short _nextexponent

                    cmp     al, $NULL_ZERO$        
                    je      _stopexponent
                    jne     _erro2                      ; unacceptable character

                    ; ---------------------
                    ;    remove spaces
                    ; ---------------------        
    _startexponent: movzx   eax, byte ptr [esi]
                    add     esi, 1

                    cmp     al, $SPACE$
                    je      short _startexponent
                         
                    ; --------------------------
                    ;     Is it a sign ?
                    ; --------------------------
                    cmp     al, $MINUS$
                    je      short @F

                    mov     ExponentSign, 0          ; is plus 
                                       
                    cmp     al, $PLUS$
                    je      short @F
                    jne     short _testzeroE                    
                    
                    ; -------------------------
                    ;   it cannot have sign
                    ;  remove zeroes or spaces
                    ; -------------------------
    _removeZE:      mov     flagZE, 1

            @@:     movzx   eax, byte ptr [esi]
                    add     esi, 1

    _testzeroE:     cmp     al, $ZERO$
                    je      short _removeZE
                    
                    cmp     al, $SPACE$
                    je      short @B                    
                    jne     short _isexponent 
            ; -------------------------------------------
            ;
            ;--------------------------------------------

    _stopinteger:   cmp     ecx, $INTEGERDIGITS_REAL4
                    jne     short _stop    
                    je      _erro1                      ; expression error
    
    _stopdecimal:   cmp     ecx, $DECIMALDIGITS_REAL4
                    jne     short _stop    
                    je      _erro1                      ; expression error

    _stopexponent:  cmp     flagZE, 1
                    je      short @F

                    cmp     ecx, $EXPONENTDIGITS_REAL4  ; number of exponent digits
                    je      _erro1                      ; expression error

            @@:     cmp     ExponentSign, 0             ; is plus 
                    je      short _stop
                    ;
                    neg     edx

                ;-----------------------------------------
                ;               adjust sign
                ;-----------------------------------------
        _stop:      mov     ecx, integerDD
                    cmp     IntegerSign, 0              ; is plus 
                    je      short @F

                    neg     ecx
                    neg     edi

                    mov     integerDD, ecx
             @@:    mov     decimalDD, edi
             
comment * ################################################################
pushad
push    ecx
print   str$(edx), " = exponent",13,10
pop     ecx
print   str$(ecx), " = integer",13,10
print   str$(ebx), " = number of decimal digits",13,10
print   str$(edi), " = decimal",13,10
popad 
          ################################################################ *
                   
                    ; -----------------
                    ; Load integer part
                    ; -----------------
                    fild    integerDD
                    
                    ; ------------------------------------
                    ;  Load decimal part even if it is 0
                    ;      If not it gives error 7
                    ; ------------------------------------
                    fild    decimalDD
                    
                    ; -------------------------------------
                    ;  multiply the decimal part by 10^-ebx
                    ;      and muliply all by 10^edx 
                    ; -------------------------------------
                    ADDDECIMALPOWERDD ebx, edx
            
                    ; --------------------------
                    ; Test for invalid operation
                    ; --------------------------
        _finish0:   fstsw   ax
                    fwait
                    shr     al, 1
                    jc      _erro7

                    mov     eax, pOut4
                    fstp    dword ptr [eax]

                    ; -------------
                    ; examine real4
                    ; -------------
                    EXAMINE2REAL4

        _finish:    xor     eax, eax
                    clc
        _exit:      pop     edi
                    pop     esi
                    pop     ebx
                    ret
                    
                    ; ---------------------------
                    ;     expression error
                    ; ---------------------------
        _erro1:     mov     eax, 1
                    mov     edx, offset _StrError1
        
        _error:     ;print   str$(eax)," ERROR eax ",13,10

                    stc
                    jmp     _exit
                    
                    ; ----------------------
                    ; unacceptable character
                    ; ----------------------
        _erro2:     mov     eax, 2
                    mov     edx, offset _StrError2
                    jmp     _error

                    ERRORS3TO9
ConvertString4D     endp
; 
; Input:
;           pReal4 = pointer
;                                
;                                                           dd 0    <------- for lenght
;           pRcl   --------------> buffer shoud be  _buffer db x dup (?), 0
;
ConvertFloat4DF     proc     pReal4:DWORD, pRcl:DWORD
                    LOCAL    expscale     :DWORD    
                    LOCAL    exponent     :DWORD
                    LOCAL    DWinteger    :DWORD
                    LOCAL    oldcw        :WORD
                    LOCAL    truncw       :WORD
                    LOCAL    string[22]   :BYTE
                                       
                    push     ebx
                    push     esi
                    push     edi                    
                    
                    mov      edi, pRcl     
                    mov      ebx, pReal4

                    ; --------------------------
                    ;       examine REAL4
                    ; --------------------------
                    mov     eax, dword ptr [ebx]                    ; EXAMINE2REAL4                                                
                    mov     edx, eax
                    shr     edx, 23                                 ; exponent and sign in DX: DH=sign (bit 0), DL=exponent

                    and     eax, 07FFFFFH                           ; fraction in EAX  <<<<<<<<<<<<<<<<<<< NEW HERE
                                        
                    cmp     dl, 0FFh        ; FFh => NAN or INFINITY
                    jne     short _notNAN   ; not NAN or infinity                    

                    ; -----------------------
                    ;  It is NAN or INFINITY
                    ; -----------------------
                    or      eax, eax
                    jnz     _erro1 
                                       
                    ;----------------------------------------------
                    ;   exponent =FFH and fraction =0 => INFINITY
                    ;----------------------------------------------
                    mov     al, '+'
    
                    shr     dh, 1
                    jnc     short @F        ; is 0 => +infinity

                    ;------------
                    ; is -infinity
                    ; ------------
                    mov     al, '-'                    
                   
            @@:     mov    byte ptr [edi], al
                    mov    dword ptr [edi+1], "IFNI"
                    mov    dword ptr [edi+5], "YTIN"
                    ;
                    mov    eax, 9
                    jmp    _exit0
                   
                    ;--------------------------
                    ; value to be converted = 0
                    ;--------------------------                                      
        _iszero:    mov      word ptr [edi], 3020h
                    mov      eax, 2
                    jmp      _exit0

                    ;------------------------------------------------
                    ;           It is not NAN or INFINITY
                    ; DL  = exponent
                    ; DH  = sign ( bit 0 )
                    ; EAX = fraction bits
                    ;------------------------------------------------
       _notNAN:     test    dl, dl
                    jnz     short _start0       ; exponent not FFh and not 0

                    ;---------------------------------------------
                    ; exponent=0, if fraction=0 => it is +0 or -0
                    ;---------------------------------------------
                    or      eax, eax
                    jz      short _iszero
                    jnz     _erro1              ; denormalized

                    ;------------------------------------------------
                    ;                 START HERE
                    ; DH  = sign ( bit 0 )           test for sign                    
                    ;------------------------------------------------
       _start0:     shr     dh, 1  
                    jnc     short @F            ; is 0 => is POSITIVE,  1 => is NEGATIVE
                                       

                    mov     byte ptr [edi], '-'
                    add     edi, 1
                    
                    ; ---------------------
                    ; start load the real 4
                    ; ---------------------
            @@:     ; finit
                    fld      dword ptr [ebx]        ; st(0) = X
                                  
                    ;--------------------------------------
                    ;           get the exponent
                    ; exponent = characteristic of log10(X)
                    ;--------------------------------------
                    fldlg2                  ; st(0)=log10(2) ; st(1)=X
                    fld      st(1)          ; copy st(1)=X to st(0)
                    fabs                    ; positive value
                    fyl2x                   ; log10(2)*log2(X) = log10(X)

                    ; ---------------------------------
                    ; X          in st(2) st(1)
                    ; log10(2)   in st(1) removed fyl2x
                    ; |X|        in st(0) removed fyl2x
                    ; log10(|X|) in st(0) st(0)    
                    ; ---------------------------------                                            

                    ; -----------------------------------------------------------
                    ;       Convert the logarithm to a truncated integer
                    ; store the truncating value of the logarithm and doesnt POP
                    ; -----------------------------------------------------------      
                    SETTRUNCATE oldcw, truncw
                    fistp     exponent
                    fldcw     oldcw                             ; load back the former control word
                          
                    ; -------------------------
                    ; Get exponent scale factor
                    ; -------------------------
                    mov     edx, $DECIMALPLACES_REAL4           ; 6 decimal places
                    sub     edx, exponent
                    mov     expscale, edx
           
                    mov     ebx, offset _PowerTblDQP38
        
                    or      edx, edx 
                    jns     short @F

                    neg     edx
                    mov     ebx, offset _PowerTblDQM38
            
                    cmp     edx, 46
                    jbe     short _load0
                    
                    ; -----------------------------------
                    ;   multiply st(0) by 10^expscale
                    ; -----------------------------------
    _power10:       fild    expscale
                    fldl2t             ; log2(10)
                    fmul               ; log2(10)*exponent
                    fld     st
                    frndint            ; get the characteristic
                    fxch
                    fsub    st,st(1)   ; get only the mantissa but keep the characteristic
                    f2xm1              ; 2^(mantissa)-1
                    fld1
                    fadd               ; add 1 and pop
                    fscale
                    fstp    st(1)      ; remove the characteristic
                    jmp     _multiply

            @@:     cmp     edx, 46
                    ja      short _power10
            
        _load0:     fld     qword ptr [ebx+edx*8]
                                        
                    ; -------------------------                    
                    ; Multiply X by 10^expscale
                    ; -------------------------
    _multiply:      fmul

                    ; -----------------
                    ; store the integer
                    ; -----------------
                    fistp   dword ptr DWinteger
                    fstsw   ax
                    fwait
                    shr     ax, 1              ; test for invalid operation
                    jc      _erro2  

                    ; -------------------------
                    ; convert integer to string
                    ; -------------------------
                    mov     eax, dword ptr DWinteger

                    or      eax, eax
                    jns     short @F
                    
                    neg     eax
                                        
            @@:     lea     esi, string+20
                    xor	    ebx, ebx
                                
                    ; ----------------------------------------
                    ; Divide EAX by 10 and get the digit in CL
                    ; ----------------------------------------
            @@:     DIVISION2BY10
                    ;
                    sub     esi, 1                
                    mov     byte ptr [esi], cl              ; cl  = digit
                    add	    ebx, 1					; ebx = number of digits
                    ;
                    ; EDX:EAX = 0 ?
                    ; -------------
                    or      eax, eax
                    jnz	    short @B
                    
                    mov     edx, $DECIMALPLACES_REAL4
                    mov     eax, exponent

                    cmp     edx, ebx
                    jne     short @F

                    ; --------------------------
                    ;   Correct the exponent
                    ; --------------------------
                    sub     eax, 1
                    mov     exponent, eax

                    ; ------------------------------------------
                    ; Regular notation or scientific notation ?
                    ; ebx = total number of digits in the buffer
                    ; edx = number of decimal places
                    ; ecx = exponent
                    ; ------------------------------------------
            @@:     or      eax, eax
                    jns     short _ispositive       ; exponent is positive
                    mov     ecx, eax
                    neg     ecx

                    ;cmp     ecx, 3  
                    cmp     ecx, edx
                    ja      _isscientific

                    ; --------------------------
                    ; Regular notation - decimal
                    ; --------------------------
                    mov     word ptr [edi], '.0'
                    add     edi, 2
                    
                    ; ------------------------------
                    ; If exponent = -1 it is 0.ddddd
                    ; If exponent = -2 it is 0.0dddd
                    ;         and so on
                    ; ------------------------------
                    sub     ecx, 1
                    jz      _movedecimal
                    
                    sub     edx, ecx
                    
                    ; -------------------------
                    ;        insert 0
                    ; -------------------------
            @@:     mov     byte ptr [edi], '0'
                    add     edi, 1
                    sub     ecx, 1
                    jnz     short @B

    _movedecimal:   movzx   eax, byte ptr [esi]
                    mov     byte ptr [edi], al
                    add     edi, 1
                    add     esi, 1

                    sub     edx, 1
                    jnz     short _movedecimal
                    jz      short _removeD
                    ; ------------------
                    ; remove last zeroes
                    ; ------------------
            @@:     sub     edi, 1
        _removeD:   movzx   eax, word ptr [edi-2]
                    cmp     al, '.'
                    je      _finish
                    ;
                    cmp     ah, '0'
                    je      short @B
                    jmp     _finish

            ;--------------------------------------------------
            ; ebx = total number of digits
            ; edx = number of decimal places
            ; eax = exponent
            ;--------------------------------------------------
                    ; -------------------
                    ;   exponent >= 0
                    ; -------------------
    _ispositive:    cmp     eax, edx
                    jae     _isscientific
                                   
                    ; --------------------------
                    ; Regular notation - integer
                    ; --------------------------
                    add     eax, 1              ; number of integer digits
                    sub     ebx, eax            ; number of decimal digits
                    ;
            @@:     movzx   edx, byte ptr [esi]
                    mov     byte ptr [edi], dl
                    add     edi, 1
                    add     esi, 1                    

                    sub     eax, 1
                    jnz     short @B
                    ;
                    or      ebx, ebx
                    jz      _finish
                    
                    mov     byte ptr [edi], '.'
                    add     edi, 1
                    ;
            @@:     movzx   eax, byte ptr [esi]
                    mov     byte ptr [edi], al
                    add     edi, 1
                    add     esi, 1                    

                    sub     ebx, 1
                    jnz     short @B
                    jmp     short _removeI
                    ; ------------------
                    ; remove last zeroes
                    ; ------------------
            @@:     sub     edi, 1
        _removeI:   movzx   eax, word ptr [edi-2]
                    cmp     al, '.'
                    je      _finish
                    ;
                    cmp     ah, '0'
                    je      short @B
                    jmp     _finish
                                        
            ;--------------------------------------------------
            ;                 is Scientific
            ;--------------------------------------------------
    _isscientific:  sub     ebx, 1
                    
                    movzx   edx, byte ptr [esi]
                    mov     dh, '.'                    
                    add     esi, 1
                    ;
                    mov     word ptr [edi], dx
                    add     edi, 2
                    
            @@:     movzx   edx, byte ptr [esi]
                    mov     byte ptr [edi], dl
                    ;
                    add     edi, 1
                    add     esi, 1

                    sub     ebx, 1
                    jnz     short @B
                    jz      short _removeE
                    
                    ; ------------------
                    ; remove last zeroes
                    ; ------------------
            @@:     sub     edi, 1
        _removeE:   movzx   edx, word ptr [edi-2]
                    cmp     dl, '.'
                    je      short @F
                    ;
                    cmp     dh, '0'
                    je      short @B

            @@:     EXPONENTREAL_4R
                            
        _finish:    mov      eax, edi
                    mov      edi, pRcl
                    sub      eax, edi

        _exit0:     mov      dword ptr [edi-4], eax
                    mov      byte ptr [edi+eax], 0
                                              
                    xor      eax, eax
                    clc
                    
        _exit:      pop      edi
                    pop      esi
                    pop      ebx
                    ret
                    
                    ; --------------------
                    ; is indefinite or NAN
                    ; --------------------
        _erro1:     mov    eax, 1
                    ;
        _error:     mov    dword ptr [edi+0], "ORRE"
                    mov    byte ptr [edi+4], "R"
                    ;
                    mov      dword ptr [edi-4], 5
                    mov      byte ptr [edi+5], 0
                    stc
                    jmp    _exit
                                        
                    ; -----------------
                    ; invalid operation
                    ; -----------------
        _erro2:     fclex
                    mov      eax, 2
                    stc 
                    jmp      _error
ConvertFloat4DF     endp
; 
; Input:
;           pReal4 = pointer
;                                
;           pString   --------------> buffer shoud be  _buffer db x dup (?), 0
;
ConvertReal4DF      proc     pReal4:DWORD, pString:DWORD
                    LOCAL    expscale     :DWORD    
                    LOCAL    exponent     :DWORD
                    LOCAL    DWinteger    :DWORD
                    LOCAL    oldcw        :WORD
                    LOCAL    truncw       :WORD
                    LOCAL    string[22]   :BYTE
                                       
                    push     ebx
                    push     esi
                    push     edi                    
                    
                    mov      edi, pString     
                    mov      ebx, pReal4
                    
                    ; --------------------------
                    ;       examine REAL4
                    ; --------------------------
                    mov     eax, dword ptr [ebx]                    ; EXAMINE2REAL4                                                
                    mov     edx, eax
                    shr     edx, 23                                 ; exponent and sign in DX: DH=sign (bit 0), DL=exponent

                    and     eax, 07FFFFFH                           ; fraction in EAX  <<<<<<<<<<<<<<<<<<< NEW HERE
                                        
                    cmp     dl, 0FFh        ; FFh => NAN or INFINITY
                    jne     short _notNAN   ; not NAN or infinity                    

                    ; -----------------------
                    ;  It is NAN or INFINITY
                    ; -----------------------
                    or      eax, eax
                    jnz     _erro1 
                                       
                    ;----------------------------------------------
                    ;   exponent =FFH and fraction =0 => INFINITY
                    ;----------------------------------------------
                    mov     al, '+'
    
                    shr     dh, 1
                    jnc     short @F        ; is 0 => +infinity

                    ;------------
                    ; is -infinity
                    ; ------------
                    mov     al, '-'                    
                   
            @@:     mov    byte ptr [edi], al
                    mov    dword ptr [edi+1], "IFNI"
                    mov    dword ptr [edi+5], "YTIN"
                    ;
                    mov    eax, 9
                    jmp    _exit0
                   
                    ;--------------------------
                    ; value to be converted = 0
                    ;--------------------------                                      
        _iszero:    mov      word ptr [edi], 3020h
                    mov      eax, 2
                    jmp      _exit0

                    ;------------------------------------------------
                    ;           It is not NAN or INFINITY
                    ; DL  = exponent
                    ; DH  = sign ( bit 0 )
                    ; EAX = fraction bits
                    ;------------------------------------------------
       _notNAN:     test    dl, dl
                    jnz     short _start0       ; exponent not FFh and not 0

                    ;---------------------------------------------
                    ; exponent=0, if fraction=0 => it is +0 or -0
                    ;---------------------------------------------
                    or      eax, eax
                    jz      short _iszero
                    jnz     _erro1              ; denormalized

                    ;------------------------------------------------
                    ;                 START HERE
                    ; DH  = sign ( bit 0 )           test for sign                    
                    ;------------------------------------------------
       _start0:     shr     dh, 1  
                    jnc     short @F            ; is 0 => is POSITIVE,  1 => is NEGATIVE
                                       

                    mov     byte ptr [edi], '-'
                    add     edi, 1
                    
                    ; ---------------------
                    ; start load the real 4
                    ; ---------------------
            @@:     ; finit
                    fld      dword ptr [ebx]        ; st(0) = X
                                  
                    ;--------------------------------------
                    ;           get the exponent
                    ; exponent = characteristic of log10(X)
                    ;--------------------------------------
                    fldlg2                  ; st(0)=log10(2) ; st(1)=X
                    fld      st(1)          ; copy st(1)=X to st(0)
                    fabs                    ; positive value
                    fyl2x                   ; log10(2)*log2(X) = log10(X)

                    ; ---------------------------------
                    ; X          in st(2) st(1)
                    ; log10(2)   in st(1) removed fyl2x
                    ; |X|        in st(0) removed fyl2x
                    ; log10(|X|) in st(0) st(0)    
                    ; ---------------------------------                                            

                    ; -----------------------------------------------------------
                    ;       Convert the logarithm to a truncated integer
                    ; store the truncating value of the logarithm and doesnt POP
                    ; -----------------------------------------------------------      
                    SETTRUNCATE oldcw, truncw
                    fistp     exponent
                    fldcw     oldcw                             ; load back the former control word
                          
                    ; -------------------------
                    ; Get exponent scale factor
                    ; -------------------------
                    mov     edx, $DECIMALPLACES_REAL4           ; 6 decimal places
                    sub     edx, exponent
                    mov     expscale, edx
           
                    mov     ebx, offset _PowerTblDQP38
        
                    or      edx, edx 
                    jns     short @F

                    neg     edx
                    mov     ebx, offset _PowerTblDQM38
            
                    cmp     edx, 46
                    jbe     short _load0
                    
                    ; -----------------------------------
                    ;   multiply st(0) by 10^expscale
                    ; -----------------------------------
    _power10:       fild    expscale
                    fldl2t             ; log2(10)
                    fmul               ; log2(10)*exponent
                    fld     st
                    frndint            ; get the characteristic
                    fxch
                    fsub    st,st(1)   ; get only the mantissa but keep the characteristic
                    f2xm1              ; 2^(mantissa)-1
                    fld1
                    fadd               ; add 1 and pop
                    fscale
                    fstp    st(1)      ; remove the characteristic
                    jmp     _multiply

            @@:     cmp     edx, 46
                    ja      short _power10
            
        _load0:     fld     qword ptr [ebx+edx*8]
                                        
                    ; -------------------------                    
                    ; Multiply X by 10^expscale
                    ; -------------------------
    _multiply:      fmul

                    ; -----------------
                    ; store the integer
                    ; -----------------
                    fistp   dword ptr DWinteger
                    fstsw   ax
                    fwait
                    shr     ax, 1              ; test for invalid operation
                    jc      _erro2  

                    ; -------------------------
                    ; convert integer to string
                    ; -------------------------
                    mov     eax, dword ptr DWinteger

                    or      eax, eax
                    jns     short @F
                    
                    neg     eax
                                        
            @@:     lea     esi, string+20
                    xor	    ebx, ebx
                                
                    ; ----------------------------------------
                    ; Divide EAX by 10 and get the digit in CL
                    ; ----------------------------------------
            @@:     DIVISION2BY10
                    ;
                    sub     esi, 1                
                    mov     byte ptr [esi], cl              ; cl  = digit
                    add	    ebx, 1					; ebx = number of digits
                    ;
                    ; EDX:EAX = 0 ?
                    ; -------------
                    or      eax, eax
                    jnz	    short @B
                    
                    mov     edx, $DECIMALPLACES_REAL4
                    mov     eax, exponent

                    cmp     edx, ebx
                    jne     short @F

                    ; --------------------------
                    ;   Correct the exponent
                    ; --------------------------
                    sub     eax, 1
                    mov     exponent, eax

                    ; ------------------------------------------
                    ; Regular notation or scientific notation ?
                    ; ebx = total number of digits in the buffer
                    ; edx = number of decimal places
                    ; ecx = exponent
                    ; ------------------------------------------
            @@:     or      eax, eax
                    jns     short _ispositive       ; exponent is positive
                    mov     ecx, eax
                    neg     ecx

                    ;cmp     ecx, 3  
                    cmp     ecx, edx
                    ja      _isscientific

                    ; --------------------------
                    ; Regular notation - decimal
                    ; --------------------------
                    mov     word ptr [edi], '.0'
                    add     edi, 2
                    
                    ; ------------------------------
                    ; If exponent = -1 it is 0.ddddd
                    ; If exponent = -2 it is 0.0dddd
                    ;         and so on
                    ; ------------------------------
                    sub     ecx, 1
                    jz      _movedecimal
                    
                    sub     edx, ecx
                    
                    ; -------------------------
                    ;        insert 0
                    ; -------------------------
            @@:     mov     byte ptr [edi], '0'
                    add     edi, 1
                    sub     ecx, 1
                    jnz     short @B

    _movedecimal:   movzx   eax, byte ptr [esi]
                    mov     byte ptr [edi], al
                    add     edi, 1
                    add     esi, 1

                    sub     edx, 1
                    jnz     short _movedecimal
                    jz      short _removeD
                    ; ------------------
                    ; remove last zeroes
                    ; ------------------
            @@:     sub     edi, 1
        _removeD:   movzx   eax, word ptr [edi-2]
                    cmp     al, '.'
                    je      _finish
                    ;
                    cmp     ah, '0'
                    je      short @B
                    jmp     _finish

            ;--------------------------------------------------
            ; ebx = total number of digits
            ; edx = number of decimal places
            ; eax = exponent
            ;--------------------------------------------------
                    ; -------------------
                    ;   exponent >= 0
                    ; -------------------
    _ispositive:    cmp     eax, edx
                    jae     _isscientific
                                   
                    ; --------------------------
                    ; Regular notation - integer
                    ; --------------------------
                    add     eax, 1              ; number of integer digits
                    sub     ebx, eax            ; number of decimal digits
                    ;
            @@:     movzx   edx, byte ptr [esi]
                    mov     byte ptr [edi], dl
                    add     edi, 1
                    add     esi, 1                    

                    sub     eax, 1
                    jnz     short @B
                    ;
                    or      ebx, ebx
                    jz      _finish
                    
                    mov     byte ptr [edi], '.'
                    add     edi, 1
                    ;
            @@:     movzx   eax, byte ptr [esi]
                    mov     byte ptr [edi], al
                    add     edi, 1
                    add     esi, 1                    

                    sub     ebx, 1
                    jnz     short @B
                    jmp     short _removeI
                    ; ------------------
                    ; remove last zeroes
                    ; ------------------
            @@:     sub     edi, 1
        _removeI:   movzx   eax, word ptr [edi-2]
                    cmp     al, '.'
                    je      _finish
                    ;
                    cmp     ah, '0'
                    je      short @B
                    jmp     _finish
                                        
            ;--------------------------------------------------
            ;                 is Scientific
            ;--------------------------------------------------
    _isscientific:  sub     ebx, 1
                    
                    movzx   edx, byte ptr [esi]
                    mov     dh, '.'                    
                    add     esi, 1
                    ;
                    mov     word ptr [edi], dx
                    add     edi, 2
                    
            @@:     movzx   edx, byte ptr [esi]
                    mov     byte ptr [edi], dl
                    ;
                    add     edi, 1
                    add     esi, 1

                    sub     ebx, 1
                    jnz     short @B
                    jz      short _removeE
                    
                    ; ------------------
                    ; remove last zeroes
                    ; ------------------
            @@:     sub     edi, 1
        _removeE:   movzx   edx, word ptr [edi-2]
                    cmp     dl, '.'
                    je      short @F
                    ;
                    cmp     dh, '0'
                    je      short @B

            @@:     EXPONENTREAL_4R
                            
        _finish:    mov      eax, edi
                    mov      edi, pString
                    sub      eax, edi

        _exit0:     ; mov      dword ptr [edi-4], eax   ; doest use length
                    mov      byte ptr [edi+eax], 0
                                              
                    xor      eax, eax
                    clc
                    
        _exit:      pop      edi
                    pop      esi
                    pop      ebx
                    ret
                    
                    ; --------------------
                    ; is indefinite or NAN
                    ; --------------------
        _erro1:     mov    eax, 1
                    ;
        _error:     mov    dword ptr [edi+0], "ORRE"
                    mov    byte ptr [edi+4], "R"
                    ;
                    ;mov      dword ptr [edi-4], 5      ; doest use length
                    mov      byte ptr [edi+5], 0
                    stc
                    jmp    _exit
                                        
                    ; -----------------
                    ; invalid operation
                    ; -----------------
        _erro2:     fclex
                    mov      eax, 2
                    stc 
                    jmp      _error
ConvertReal4DF      endp
; 
